#include "connect4.hpp"
#include <sstream>

using namespace std;

namespace game
{
  static vector<vector<uint64_t>> create_hash_values()
  {
    default_random_engine generator;
    uniform_int_distribution<uint64_t> distribution;
    vector<vector<uint64_t>> res(7, vector<uint64_t>(6, 0));
    for (int i = 0; i < 7; ++i) 
      {
	for (int j = 0; j < 6; ++j)
	  {
	    res[i][j] = distribution(generator);
	  }
      }
    return res;
  }

  vector<vector<uint64_t>> connect4::cross_hash_values = create_hash_values();
  vector<vector<uint64_t>> connect4::circle_hash_values = create_hash_values();

  connect4::connect4()
  {
  }

  shared_ptr<game<connect4_state>> connect4::do_copy() const
  {
    return shared_ptr<connect4>(new connect4(*this));
  }

  connect4_state connect4::get_state()
  {
    return state;
  }
  
  void connect4::set_state(const connect4_state& s)
  {
    state = s; 
  }

  bool connect4::end_of_game() const
  {
    return state.first_player_win || state.second_player_win || state.total_moves == 42;
  }

  bool connect4::won(std::uint8_t player) const
  {
    if (player == 0) return state.first_player_win;
    return state.second_player_win;
  }
  
  bool connect4::lost(std::uint8_t player) const
  {
    if (player == 0) return state.second_player_win;
    return state.first_player_win;
  }

  bool connect4::draw(std::uint8_t player) const
  {
    if (state.first_player_win || state.second_player_win) return false;
    return state.total_moves == 42;
  }

  uint8_t connect4::current_player() const
  {
    return state.total_moves & 1 ? CIRCLE : CROSS; 
  }

  // uint8_t connect4::current_player_representation() const
  // {
  //   return state.total_moves & 1 ? CIRCLE_REPRESENTATION : CROSS_REPRESENTATION; 
  // }

  // uint8_t connect4::player_representation_to_player(uint8_t player) const
  // {
  //   return player == CIRCLE_REPRESENTATION ? CIRCLE : (player == CROSS_REPRESENTATION ? CROSS : -1); 
  // }

  int connect4::value(uint8_t player) const
  {
    if (player == CROSS)
      {
	return state.first_player_win ? 1 : (state.second_player_win ? -1 : 0);
      }
    else if (player == CIRCLE)
      {
	return state.second_player_win ? 1 : (state.first_player_win ? -1 : 0);
      }
    return 0;
  }

  uint16_t connect4::number_of_moves() const
  {
    return state.nb_moves;
  }

  bool connect4::get(uint64_t bitboard, uint8_t col, uint8_t row) const
  {
    return bitboard & (1LL << ((col << 3) + row));
    //return (state.board[i] >> (j << 1)) & 3;
  }

#define set(bitboard, col, row) (bitboard |= (1LL << (((col) << 3) + (row))))

  // bool connect4::vertical(uint8_t position, uint8_t free, uint8_t player) const
  // {
  //   return free >= 3 && get(position, free - 1) == player && get(position, free - 2) == player && get(position, free - 3) == player;
  // }

  // bool connect4::horizontal(uint8_t position, uint8_t free, uint8_t player) const
  // {
  //   uint8_t sum = 0;
  //   if (position >= 1 && get(position - 1, free) == player)
  //     {
  // 	++sum;
  // 	if (position >= 2 && get(position - 2, free) == player)
  // 	  {
  // 	    ++sum;
  // 	    if (position >= 3 && get(position - 3, free) == player) ++sum;
  // 	  }
  //     }
  //   if (position <= 5 && get(position + 1, free) == player)
  //     {
  // 	++sum;
  // 	if (position <= 4 && get(position + 2, free) == player)
  // 	  {
  // 	    ++sum;
  // 	    if (position <= 3 && get(position + 3, free) == player) ++sum;
  // 	  }
  //     }
  //   return sum >= 3;
  // }

  // bool connect4::diagonal(uint8_t position, uint8_t free, uint8_t player) const
  // {
  //   uint8_t sum = 0;
  //   if (position >= 1 && free <= 4 && get(position - 1, free + 1) == player)
  //     {
  // 	++sum;
  // 	if (position >= 2 && free <= 3 && get(position - 2, free + 2) == player)
  // 	  {
  // 	    ++sum;
  // 	    if (position >= 3 && free <= 2 && get(position - 3, free + 3) == player) ++sum;
  // 	  }
  //     }
  //   if (position <= 5 && free >= 1 && get(position + 1, free - 1) == player)
  //     {
  // 	++sum;
  // 	if (position <= 4 && free >= 2 && get(position + 2, free - 2) == player)
  // 	  {
  // 	    ++sum;
  // 	    if (position <= 3 && free >= 3 && get(position + 3, free - 3) == player) ++sum;
  // 	  }
  //     }
  //   return sum >= 3;
  // }

  // bool connect4::other_diagonal(uint8_t position, uint8_t free, uint8_t player) const
  // {
  //   uint8_t sum = 0;
  //   if (position >= 1 && free >= 1 && get(position - 1, free - 1) == player)
  //     {
  // 	++sum;
  // 	if (position >= 2 && free >= 2 && get(position - 2, free - 2) == player)
  // 	  {
  // 	    ++sum;
  // 	    if (position >= 3 && free >= 3 && get(position - 3, free - 3) == player) ++sum;
  // 	  }
  //     }
  //   if (position <= 5 && free <= 4 && get(position + 1, free + 1) == player)
  //     {
  // 	++sum;
  // 	if (position <= 4 && free <= 3 && get(position + 2, free + 2) == player)
  // 	  {
  // 	    ++sum;
  // 	    if (position <= 3 && free <= 2 && get(position + 3, free + 3) == player) ++sum;
  // 	  }
  //     }
  //   return sum >= 3;
  // }

  // void connect4::update_win(uint8_t position, uint8_t free)
  // {
  //   uint8_t player = current_player_representation();
  //   bool win = 
  //     vertical(position, free, player)
  //     ||
  //     horizontal(position, free, player)
  //     ||
  //     diagonal(position, free, player)
  //     ||
  //     other_diagonal(position, free, player);
  //   if (win) 
  //     {
  // 	if (player == CROSS_REPRESENTATION) state.first_player_win = true;
  // 	else state.second_player_win = true;
  //     }
  // }

  void connect4::update_win()
  {
    if (has_won(state.cross_bitboard)) state.first_player_win = true;
    else if (has_won(state.circle_bitboard)) state.second_player_win = true;
  }

  bool connect4::has_won(uint64_t bitboard)
  {
    int64_t y = bitboard & (bitboard >> 7);
    if (y & (y >> 2 * 7)) // check \ diagonal
      return true;
    y = bitboard & (bitboard >> 8);
    if (y & (y >> 2 * 8)) // check horizontal -
      return true;
    y = bitboard & (bitboard >> 9);
    if (y & (y >> 2 * 9)) // check / diagonal
      return true;
    y = bitboard & (bitboard >> 1);
    if (y & (y >> 2))     // check vertical |
      return true;
    return false;
  }

  void connect4::update_moves(uint16_t move)
  {
    --state.nb_moves;
    uint32_t prefix = state.moves >> ((move + 1) * 3);
    uint8_t shift = 32 - move * 3;
    state.moves = (uint32_t)((uint64_t)state.moves << shift) >> shift;
    state.moves |= prefix << (move * 3);
  }

  void connect4::play(uint16_t m)
  {    
    uint8_t position = ((state.moves >> (m * 3)) & 7);
    uint8_t p = position * 3;
    uint8_t f = (state.free >> p) & 7;
    //    state.hash_value ^= current_player() == CIRCLE ? circle_hash_values[position][f] : cross_hash_values[position][f];
    //    state.board[position] |= current_player_representation() << (f << 1);
    if (current_player() == CROSS)
      {
	set(state.cross_bitboard, position, f);
      }
    else
      {
	set(state.circle_bitboard, position, f);
      }
    update_win();
    ++f;
    state.free = (state.free & ~(((uint32_t)7) << p)) | (f << p);
    if (f == 6)
      {
	update_moves(m);
      }
    ++state.total_moves;
  }

  string connect4::player_to_string(uint8_t player) const
  {
    return player == CROSS ? "X" : (player == CIRCLE ? "O" : " ");
  }

  string connect4::move_to_string(uint16_t m) const
  {
    uint8_t position = ((state.moves >> (m * 3)) & 7);
    return std::to_string(position);
  }

  set<int> connect4::to_input_vector() const
  {
    set<int> res;
    int k = 0;
    if (current_player()) res.insert(k++);
    for (int col = 0; col < 7; ++col) 
      {
	for (int row = 0; row < 6; ++row)
	  {
	    if (get(state.cross_bitboard, col, row)) res.insert(k++);
	  }
      }
    for (int col = 0; col < 7; ++col)
      {
	for (int row = 0; row < 6; ++row)
	  {
	    if (get(state.circle_bitboard, col, row)) res.insert(k++);
	  }
      }
    return res;
  }

  void connect4::from_input_vector(const std::set<int>& input)
  {
    state = connect4_state();
    for (int index : input) 
      {
	if (index == 1) continue;
	if (index <= 43)
	  {
	    index -= 2;
	    state.cross_bitboard |= 1LL << (index + (index / 6) * 2);
	  }
	else 
	  {
	    index -= 44;
	    state.circle_bitboard |= 1LL << (index + (index / 6) * 2);
	  }
	state.total_moves += 1;
      }
  }

  string connect4::to_string() const
  {
    stringbuf buffer;
    ostream os(&buffer);
    for (int y = 5; y >= 0; --y) 
      {
	for (int k = 0; k < 7; ++k) os << "+-";
	os << "+" << endl;
	for (int x = 0; x < 7; ++x)
	  {
	    os << "|" << (get(state.cross_bitboard, x, y) ? player_to_string(CROSS) : (get(state.circle_bitboard, x, y) ? player_to_string(CIRCLE) : " "));
	  }
	os << "|" << endl;
      }
    for (int k = 0; k < 7; ++k) os << "+-";
    os << "+" << endl;
    for (int k = 0; k < 7; ++k) os << " " << k;
    os << endl;
    return buffer.str();
  }

  // void connect4::playout(mt19937& engine, int max_depth)
  // {
  //   while (!end_of_game())
  //     {
  // 	uniform_int_distribution<uint16_t> distribution(0, number_of_moves() - 1);
  // 	uint16_t move = distribution(engine);
  // 	play(move);
  //     }
  // }

  std::uint64_t connect4::hash() const
  {
    return state.hash_value;
  }

  std::uint64_t connect4::hash(std::uint16_t m) const
  {
    uint8_t position = ((state.moves >> (m * 3)) & 7);
    uint8_t p = position * 3;
    uint8_t f = (state.free >> p) & 7;
    return state.hash_value ^ (current_player() == CIRCLE ? circle_hash_values[position][f] : cross_hash_values[position][f]);
  }
  
  ostream& operator<<(ostream& os, const connect4& c4)
  {
    os << c4.to_string() << endl;
    return os;
  }
}
